""" API for Bosch API server for Indego lawn mower """
import json
import logging
import typing

import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from requests.exceptions import RequestException
from requests.exceptions import Timeout
from requests.exceptions import TooManyRedirects

from . import __version__
from .const import COMMANDS
from .const import CONTENT_TYPE
from .const import CONTENT_TYPE_JSON
from .const import DEFAULT_BODY
from .const import DEFAULT_CALENDAR
from .const import DEFAULT_HEADER
from .const import DEFAULT_URL
from .indego_base_client import IndegoBaseClient

_LOGGER = logging.getLogger(__name__)


class IndegoClient(IndegoBaseClient):
    """Class for Indego Non-Async Client."""

    def __init__(
        self,
        username: str,
        password: str,
        serial: str,
        map_filename: str = None,
        api_url: str = DEFAULT_URL,
    ):
        """Initialize the Client

        Args:
            username (str): username for Indego Account
            password (str): password for Indego Account
            serial (str): serial number of the mower
            map_filename (str, optional): Filename to store maps in. Defaults to None.
            api_url (str, optional): url for the api, defaults to DEFAULT_URL.
            
        """
        super().__init__(username, password, serial, map_filename, api_url)

    def __enter__(self):
        """Enter for with."""
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        """Exit for with."""
        pass

    def start(self):
        """Login if not done."""
        if not self._logged_in:
            self.login()

    def update_all(self, force=False):
        """Update all states."""
        self.update_generic_data()
        self.update_state(force)
        self.update_alerts()
        self.update_last_completed_mow()
        self.update_location()
        self.update_next_mow()
        self.update_operating_data()
        self.update_updates_available()
        self.update_users()
        self.update_network()

    def update_generic_data(self):
        """Update generic data."""
        self._update_generic_data(self.get(f"alms/{self._serial}"))

    def update_alerts(self):
        """Update alerts."""
        self._update_alerts(self.get("alerts"))

    def update_last_completed_mow(self):
        """Update last completed mow."""
        self._update_last_completed_mow(
            self.get(f"alms/{self._serial}/predictive/lastcutting")
        )

    def update_location(self):
        """Update location."""
        self._update_location(self.get(f"alms/{self._serial}/predictive/location"))

    def update_next_mow(self):
        """Update next mow datetime."""
        self._update_next_mow(self.get(f"alms/{self._serial}/predictive/nextcutting"))

    def update_operating_data(self):
        """Update operating data."""
        self._update_operating_data(self.get(f"alms/{self._serial}/operatingData"))

    def update_state(self, force=False, longpoll=False, longpoll_timeout=120):
        """Update state. Can be both forced and with longpoll.

        Args:
            force (bool, optional): Force the state refresh, wakes up the mower. Defaults to False.
            longpoll (bool, optional): Do a longpoll. Defaults to False.
            longpoll_timeout (int, optional): Timeout of the longpoll. Defaults to 120.

        """
        path = f"alms/{self._serial}/state"
        # _LOGGER.debug("---Update State")
        if longpoll:
            if self.state:
                last_state = self.state.state
            else:
                last_state = 0
            path = f"{path}?longpoll=true&timeout={longpoll_timeout}&last={last_state}"
        if force:
            path = f"{path}?forceRefresh=true"

        self._update_state(self.get(path))

    def update_updates_available(self):
        """Update updates available."""
        if self._online:
            self._update_updates_available(self.get(f"alms/{self._serial}/updates"))

    def update_users(self):
        """Update users."""
        self._update_users(self.get(f"users/{self._userid}"))

    def update_network(self):
        """Update network."""
        self._update_network(self.get(f"alms/{self._serial}/network"))

    def download_map(self, filename: str = None):
        """Download the map.

        Args:
            filename (str, optional): Filename for the map. Defaults to None, can also be filled by the filename set in init.

        """
        if filename:
            self.map_filename = filename
        if not self.map_filename:
            _LOGGER.error("No map filename defined.")
            return
        map = self.get(f"alms/{self._serial}/map")
        if map:
            with open(self.map_filename, "wb") as afp:
                afp.write(map)

    def put_command(self, command: str):
        """Send a command to the mower.

        Args:
            command (str): command should be one of "mow", "pause", "returnToDock"

        Returns:
            str: either result of the call or 'Wrong Command'

        """
        if command in COMMANDS:
            return self.put(f"alms/{self._serial}/state", {"state": command})
        _LOGGER.warning("%s not valid", command)
        return "Wrong Command!"

    def put_mow_mode(self, command: typing.Any):
        """Set the mower to mode manual (false-ish) or predictive (true-ish).

        Args:
            command (str/bool): should be str that is bool-ish (true, True, false, False) or a bool.

        Returns:
            str: either result of the call or 'Wrong Command'
            
        """
        if command in ("true", "false", "True", "False") or isinstance(command, bool):
            return self.put(f"alms/{self._serial}/predictive", {"enabled": command})
        _LOGGER.warning("%s not valid", command)
        return "Wrong Command!"

    def put_predictive_cal(self, calendar: dict = DEFAULT_CALENDAR):
        """Set the predictive calendar."""
        return self.put(f"alms/{self._serial}/predictive/calendar", calendar)

    def login(self):
        """Login to the api and store the context."""
        response = requests.post(
            f"{self._api_url}authenticate",
            json=DEFAULT_BODY,
            headers=DEFAULT_HEADER,
            auth=HTTPBasicAuth(self._username, self._password),
            timeout=30,
        )
        response.raise_for_status()
        self._login(response.json())

    def get(self, path, timeout=30):
        """Send a GET request and return the response as a dict."""
        if not self._logged_in:
            _LOGGER.warning("Please log in before calling updates.")
            return None
        url = self._api_url + path
        headers = DEFAULT_HEADER.copy()
        headers["x-im-context-id"] = self._contextid

        attempts = 0
        while attempts < 5:
            attempts = attempts + 1
            try:
                response = requests.get(url, headers=headers, timeout=timeout)
                response.raise_for_status()
                if CONTENT_TYPE_JSON in response.headers[CONTENT_TYPE].split(";"):
                    return response.json()
                else:
                    return response.content
            except (Timeout, TooManyRedirects, RequestException) as e:
                _LOGGER.error("Failed to update Indego status. %s", e)
                return None
            except Exception as e:
                _LOGGER.error("Get gave a unhandled error: %s", e)
                return None
            else:
                _LOGGER.debug("      HTTP Status code: " + str(response.status_code))
                if response.status_code == 400:
                    _LOGGER.error("      server answer: Bad Request")
                    return None
                if response.status_code == 500:
                    # _LOGGER.error("      Server answer: Internal Server Error")
                    _LOGGER.info("      Server answer: Internal Server Error")
                    return None
                if response.status_code == 501:
                    # _LOGGER.error("      Server answer: not implemented yet")
                    _LOGGER.info("      Server answer: not implemented yet")
                    return None
                if response.status_code == 504:
                    _LOGGER.info(
                        "      Server backend did not have an update before timeout"
                    )
                    return None
                if response.status_code == 204:
                    _LOGGER.info("      No content in response from server")
                    return None

                elif response.status != 200:
                    # relogin for other codes
                    _LOGGER.debug("      Try to login again")
                    self.login()
                    headers["x-im-context-id"] = self._contextid
                    continue
                else:
                    _LOGGER.debug("      Json:" + str(response.json()))
                    response.raise_for_status()
                    _LOGGER.debug("   --- GET: end")
                    answer = True
                    return response.json()

        if attempts >= 5:
            _LOGGER.warning("Tried 5 times to get data but did not succeed")
            _LOGGER.warning(
                "Do you initilize with the correct username, password and serial?"
            )
            return None

    def put(self, path, data):
        """Send a PUT request and return the response as a dict."""
        headers = DEFAULT_HEADER.copy()
        headers["x-im-context-id"] = self._contextid

        full_url = self._api_url + path
        try:
            response = requests.put(full_url, headers=headers, json=data, timeout=30)
            response.raise_for_status()
            return True
        except (Timeout, TooManyRedirects, RequestException) as e:
            _LOGGER.error("Failed to update Indego status. %s", e)
            return None
        except Exception as e:
            _LOGGER.error("Get gave a unhandled error: %s", e)
            return None



""" API for Bosch API server for Indego lawn mower """
import asyncio
import json
import logging
import typing

import aiohttp
import requests

# from aiofile import AIOFile
from aiohttp import ClientResponseError
from aiohttp import ServerTimeoutError
from aiohttp import TooManyRedirects
from aiohttp.helpers import BasicAuth

from . import __version__
from .const import COMMANDS
from .const import CONTENT_TYPE_JSON
from .const import DEFAULT_BODY
from .const import DEFAULT_CALENDAR
from .const import DEFAULT_HEADER
from .const import DEFAULT_URL
from .indego_base_client import IndegoBaseClient

_LOGGER = logging.getLogger(__name__)


class IndegoAsyncClient(IndegoBaseClient):
    """Class for Indego Async Client."""

    def __init__(
        self,
        username: str,
        password: str,
        serial: str,
        map_filename: str = None,
        api_url: str = DEFAULT_URL,
    ):
        """Initialize the Async Client

        Args:
            username (str): username for Indego Account
            password (str): password for Indego Account
            serial (str): serial number of the mower
            map_filename (str, optional): Filename to store maps in. Defaults to None.
            api_url (str, optional): url for the api, defaults to DEFAULT_URL.
            
        """
        super().__init__(username, password, serial, map_filename, api_url)
        self._session = aiohttp.ClientSession(raise_for_status=True)

    async def __aenter__(self):
        """Enter for async with."""
        await self.start()
        return self

    async def __aexit__(self, type, value, traceback):
        """Exit for async with."""
        await self.close()

    async def start(self):
        """Login if not done."""
        if not self._logged_in:
            await self.login()

    async def close(self):
        """Close the aiohttp session."""
        await self._session.close()

    async def update_all(self, force=False):
        """Update all states."""
        results = await asyncio.gather(
            *[
                self.update_generic_data(),
                self.update_alerts(),
                self.update_last_completed_mow(),
                self.update_location(),
                self.update_next_mow(),
                self.update_operating_data(),
                self.update_state(force),
                self.update_updates_available(),
                self.update_users(),
                self.update_network(),
            ],
            return_exceptions=True,
        )
        for res in results:
            if res:
                _LOGGER.warning(res)

    async def update_generic_data(self):
        """Update generic data."""
        self._update_generic_data(await self.get(f"alms/{self._serial}"))

    async def update_alerts(self):
        """Update alerts."""
        self._update_alerts(await self.get("alerts"))

    async def update_last_completed_mow(self):
        """Update last completed mow."""
        self._update_last_completed_mow(
            await self.get(f"alms/{self._serial}/predictive/lastcutting")
        )

    async def update_location(self):
        """Update location."""
        self._update_location(
            await self.get(f"alms/{self._serial}/predictive/location")
        )

    async def update_next_mow(self):
        """Update next mow datetime."""
        self._update_next_mow(
            await self.get(f"alms/{self._serial}/predictive/nextcutting")
        )

    async def update_operating_data(self):
        """Update operating data."""
        self._update_operating_data(
            await self.get(f"alms/{self._serial}/operatingData")
        )

    async def update_state(self, force=False, longpoll=False, longpoll_timeout=120):
        """Update state. Can be both forced and with longpoll.

        Args:
            force (bool, optional): Force the state refresh, wakes up the mower. Defaults to False.
            longpoll (bool, optional): Do a longpoll. Defaults to False.
            longpoll_timeout (int, optional): Timeout of the longpoll. Defaults to 120.

        """
        path = f"alms/{self._serial}/state"
        if longpoll:
            if self.state:
                last_state = self.state.state
            else:
                last_state = 0
            path = f"{path}?longpoll=true&timeout={longpoll_timeout}&last={last_state}"
        if force:
            if longpoll:
                path = f"{path}&forceRefresh=true"
            else:
                path = f"{path}?forceRefresh=true"

        self._update_state(await self.get(path))

    async def update_updates_available(self):
        """Update updates available."""
        if self._online:
            self._update_updates_available(
                await self.get(f"alms/{self._serial}/updates")
            )

    async def update_users(self):
        """Update users."""
        self._update_users(await self.get(f"users/{self._userid}"))

    async def update_network(self):
        """Update network."""
        self._update_network(await self.get(f"alms/{self._serial}/network"))

    async def download_map(self, filename: str = None):
        """Download the map.

        Args:
            filename (str, optional): Filename for the map. Defaults to None, can also be filled by the filename set in init.

        """
        if filename:
            self.map_filename = filename
        if not self.map_filename:
            _LOGGER.error("No map filename defined.")
            return
        # async with AIOFile(self.map_filename, "wb") as afp:
        #    await afp.write(await self.get(f"alms/{self._serial}/map"))
        map = self.get(f"alms/{self._serial}/map")
        if map:
            with open(self.map_filename, "wb") as afp:
                afp.write(map)

    async def put_command(self, command: str):
        """Send a command to the mower.

        Args:
            command (str): command should be one of "mow", "pause", "returnToDock"

        Returns:
            str: either result of the call or 'Wrong Command'

        """
        if command in COMMANDS:
            return await self.put(f"alms/{self._serial}/state", {"state": command})
        _LOGGER.warning("%s not valid", command)
        return "Wrong Command!"

    async def put_mow_mode(self, command: typing.Any):
        """Set the mower to mode manual (false-ish) or predictive (true-ish).

        Args:
            command (str/bool): should be str that is bool-ish (true, True, false, False) or a bool.

        Returns:
            str: either result of the call or 'Wrong Command'
            
        """
        if command in ("true", "false", "True", "False") or isinstance(command, bool):
            return await self.put(
                f"alms/{self._serial}/predictive", {"enabled": command}
            )
        _LOGGER.warning("%s not valid", command)
        return "Wrong Command!"

    async def put_predictive_cal(self, calendar: dict = DEFAULT_CALENDAR):
        """Set the predictive calendar."""
        return await self.put(f"alms/{self._serial}/predictive/calendar", calendar)

    async def login(self):
        """Login to the api and store the context."""
        async with self._session.post(
            f"{self._api_url}authenticate",
            json=DEFAULT_BODY,
            headers=DEFAULT_HEADER,
            auth=BasicAuth(self._username, self._password),
            timeout=30,
        ) as self._login_session:
            self._login(await self._login_session.json())

    async def get(self, path, timeout=30):
        """Send a GET request and return the response as a dict."""
        if not self._logged_in:
            _LOGGER.warning("Please log in before calling updates.")
            return None
        url = self._api_url + path
        headers = DEFAULT_HEADER.copy()
        headers["x-im-context-id"] = self._contextid

        attempts = 0
        while attempts < 5:
            attempts = attempts + 1
            try:
                async with self._session.get(
                    url, headers=headers, timeout=timeout
                ) as response:
                    if response.content_type == CONTENT_TYPE_JSON:
                        return await response.json()
                    else:
                        return await response.content.read()
            except (ServerTimeoutError, TooManyRedirects, ClientResponseError) as e:
                _LOGGER.error("Failed to update Indego status. %s", e)
                return None
            except Exception as e:
                _LOGGER.error("Get gave a unhandled error: %s", e)
                return None

            if response.status == 504:
                _LOGGER.error("Server backend did not have an update before timeout")
                return None
            elif response.status != 200:
                # relogin for other codes
                await self.login()
                headers["x-im-context-id"] = self._contextid
                continue

        if attempts >= 5:
            _LOGGER.warning("Tried 5 times to get data but did not succeed")
            _LOGGER.warning(
                "Do you initilize with the correct username, password and serial?"
            )
            return None

    async def put(self, path, data):
        """Send a PUT request and return the response as a dict."""
        headers = DEFAULT_HEADER.copy()
        headers["x-im-context-id"] = self._contextid

        full_url = self._api_url + path
        try:
            async with self._session.put(
                full_url, headers=headers, json=data, timeout=30
            ):
                return True
        except (ServerTimeoutError, TooManyRedirects, ClientResponseError) as e:
            _LOGGER.error("Failed to update Indego status. %s", e)
            return None
        except Exception as e:
            _LOGGER.error("Get gave a unhandled error: %s", e)
            return None



# pypi
""" API for Bosch API server for Indego lawn mower """
import requests
import json
from requests.auth import HTTPBasicAuth
import logging

# from datetime import datetime

DEFAULT_URL = "https://api.indego.iot.bosch-si.com:443/api/v1/"
# CONST TAKEN FROM homeassistant.const
CONTENT_TYPE_JSON = "application/json"
# const taken from aiohttp.hdrs
CONTENT_TYPE = "Content-Type"

_LOGGER = logging.getLogger(__name__)
# _LOGGER.setLevel(logging.INFO)
# _LOGGER.setLevel(logging.WARNING)
# _LOGGER.setLevel(logging.CRITICAL)
# _LOGGER.setLevel(logging.ERROR)
# _LOGGER.setLevel(logging.DEBUG)

# _LOGGER.debug("---------------------------------")
# _LOGGER.debug("Start")

MOWER_STATE_DESCRIPTION_DETAILED = {
    "0": "Reading status",
    "257": "Charging",
    "258": "Docked",
    "259": "Docked - Software update",
    "260": "Docked",
    "261": "Docked",
    "262": "Docked - Loading map",
    "263": "Docked - Saving map",
    "513": "Mowing",
    "514": "Relocalising",
    "515": "Loading map",
    "516": "Learning lawn",
    "517": "Paused",
    "518": "Border cut",
    "519": "Idle in lawn",
    "769": "Returning to Dock",
    "770": "Returning to Dock",
    "771": "Returning to Dock - Battery low",
    "772": "Returning to dock - Calendar timeslot ended",
    "773": "Returning to dock - Battery temp range",
    "774": "Returning to dock - requested by user/app",
    "775": "Returning to dock - Lawn complete",
    "776": "Returning to dock - Relocalising",
    "1025": "Diagnostic mode",
    "1026": "End of life",
    "1281": "Software update",
    "1537": "Stuck on lawn, help needed",
    "64513": "Sleeping",
    "99999": "Offline",
    "None": "None",
}

MOWER_STATE_DESCRIPTION = {
    "0": "Docked",
    "257": "Docked",
    "258": "Docked",
    "259": "Docked",
    "260": "Docked",
    "261": "Docked",
    "262": "Docked",
    "263": "Docked",
    "513": "Mowing",
    "514": "Mowing",
    "515": "Mowing",
    "516": "Mowing",
    "517": "Mowing",
    "518": "Mowing",
    "519": "Mowing",
    "769": "Mowing",
    "770": "Mowing",
    "771": "Mowing",
    "772": "Mowing",
    "773": "Mowing",
    "774": "Mowing",
    "775": "Mowing",
    "776": "Mowing",
    "1025": "Diagnostic mode",
    "1026": "End of life",
    "1281": "Software update",
    "1537": "Stuck",
    "64513": "Docked",
    "99999": "Offline",
    "None": "None",
}


MOWER_MODEL_DESCRIPTION = {
    "3600HA2300": "Indego 1000",
    "3600HA2301": "Indego 1200",
    "3600HA2302": "Indego 1100",
    "3600HA2303": "Indego 13C",
    "3600HA2304": "Indego 10C",
    "3600HB0100": "Indego 350",
    "3600HB0101": "Indego 400",
    "3600HB0102": "Indego S+ 350",
    "3600HB0103": "Indego S+ 400",
    "3600HB0105": "Indego S+ 350 2020",
    "3600HB0106": "Indego S+ 400 2020",
    "3600HB0301": "Indego M+ 700"
    #    '3600HB0xxx': 'Indego M+ 700' missing model number
}

MOWER_MODEL_VOLTAGE = {
    "3600HA2300": {"min": "297", "max": "369"},  # Indego 1000
    "3600HA2301": {"min": "297", "max": "369"},  # Indego 1200
    "3600HA2302": {"min": "297", "max": "369"},  # Indego 1100
    "3600HA2303": {"min": "297", "max": "369"},  # Indego 13C
    "3600HA2304": {"min": "297", "max": "369"},  # Indego 10C
    "3600HB0100": {"min": "0", "max": "100"},  # Indego 350
    "3600HB0101": {"min": "0", "max": "100"},  # Indego 400
    "3600HB0102": {"min": "0", "max": "100"},  # Indego S+ 350
    "3600HB0103": {"min": "0", "max": "100"},  # Indego S+ 400
    "3600HB0105": {"min": "0", "max": "100"},  # Indego S+ 350
    "3600HB0106": {"min": "0", "max": "100"},  # Indego S+ 400
    "3600HB0301": {"min": "0", "max": "100"}  # Indego M+ 700
    #    '3600HB0xxx': {'min': '0','max': '100'}   # Indego M+ 700
}

MOWING_MODE_DESCRIPTION = {
    "smart": "SmartMowing",
    "calendar": "Calendar",
    "manual": "Manual",
}

ALERT_ERROR_CODE = {
    "104": "Stop button pushed",
    "115": "Mower is stuck",
    "149": "Mower uutside perimeter cable",
    "151": "Perimeter cable signal missing",
    "ntfy_blade_life": "Reminder blade life",
}


class IndegoAPI:
    """Wrapper for Indego's API."""

    def __init__(self, username=None, password=None, serial=None):
        """Initialize Indego API and set headers needed later."""
        _LOGGER.info("Call IndegoAPI")
        _LOGGER.debug(
            "---------------------------------------------------------------------------"
        )
        _LOGGER.info("Bosch Indego API Initialization")

        _LOGGER.info("--- This is an information!")
        _LOGGER.warning("--- This is an warning!")
        _LOGGER.critical("--- This is an critical!")
        _LOGGER.error("--- This is an error!")
        _LOGGER.debug("--- This is an debug!!")

        # Declaring variables in case that they are read before initialized
        self._api_url = DEFAULT_URL
        # Log in settings
        self._username = username
        self._password = password
        self._headers = {CONTENT_TYPE: CONTENT_TYPE_JSON}
        self.body = {
            "device": "",
            "os_type": "Android",
            "os_version": "4.0",
            "dvc_manuf": "unknown",
            "dvc_type": "unknown",
        }
        self._jsonBody = json.dumps(self.body)
        # Properties for cached values
        self._serial = serial
        self._devices = {}

        # Empty declaration
        self._mower_state = None
        self._map_update_available = None
        self._mowed = None
        self._mowmode = None
        self._xpos = None
        self._ypos = None
        self._runtime = None
        self._mapsvgcache_ts = None
        self._svg_xPos = None
        self._svg_yPos = None
        self._mower_state_description = None
        self._mower_state_description_detailed = None
        self._total_operation = None
        self._total_charge = None
        self._total_cut = None
        self._session_operation = None
        self._session_charge = None
        self._session_cut = None
        self._email = None
        self._display_name = None
        self._language = None
        self._country = None
        self._optin = None
        self._optinapp = None
        self._alm_name = None
        self._service_counter = None
        self._needs_service = None
        self._alm_mode = None
        self._bareToolnumber = None
        self._alm_firmware_version = None
        self._model_description = None
        self._battery = None
        self._garden = None
        self._hmikeys = None
        self._battery_percent = None
        self._battery_voltage = None
        self._battery_cycles = None
        self._battery_discharge = None
        self._battery_ambient_temp = None
        self._battery_temp = None
        self._firmware_available = None
        self._mowingmode_description = None
        self._battery_percent_adjusted = None
        self._alert1_id = None
        self._alert1_error = None
        self._alert1_time = None
        self._alert1_friendly_description = None
        self._alert2_id = None
        self._alert2_error = None
        self._alert2_time = None
        self._alert2_friendly_description = None
        self._alert3_id = None
        self._alert3_error = None
        self._alert3_time = None
        self._alert3_friendly_description = None
        self._online = False
        self._offline = 0
        self._last_completed_mow = None
        self._next_mow = None

        ## Logging in
        self.login()

        _LOGGER.debug("--- Indego API: end __init__")

    def login(self):
        _LOGGER.info("Indego API: call login")
        _LOGGER.debug(
            "   >>> API-call: %s", "{}{}".format(self._api_url, "authenticate")
        )
        self._login_session = requests.post(
            "{}{}".format(self._api_url, "authenticate"),
            data=self._jsonBody,
            headers=self._headers,
            auth=HTTPBasicAuth(self._username, self._password),
            timeout=30,
        )
        _LOGGER.debug("JSON Response: " + str(self._login_session.json()))

        logindata = json.loads(self._login_session.content)
        self._contextid = logindata["contextId"]
        _LOGGER.debug("self._contextid: " + self._contextid)
        self._userid = logindata["userId"]
        _LOGGER.debug("self._userId: " + logindata["userId"])
        _LOGGER.debug("self._serial: " + self._serial)
        _LOGGER.debug("--- Indego API: end login")
        # self.initial_update()

    def show_vars(self):
        # For debug: writing all vars to log
        _LOGGER.info("Call show_vars")

        _LOGGER.debug("-Catched from authenticate")
        _LOGGER.debug(f"self._contextid: {self._contextid}")
        _LOGGER.debug(f"self._userid: {self._userid}")
        _LOGGER.debug(f"self._serial: {self._serial}")

        _LOGGER.debug("-self.getState")
        _LOGGER.debug(f"self._mower_state: {self._mower_state}")
        _LOGGER.debug(f"self._map_update_available: {self._map_update_available}")
        _LOGGER.debug(f"self._mowed: {self._mowed}")
        _LOGGER.debug(f"self._mowmode: {self._mowmode}")
        _LOGGER.debug(f"self._xPos: {self._xpos}")
        _LOGGER.debug(f"self._yPos: {self._ypos}")
        _LOGGER.debug(f"self._runtime: {self._runtime}")
        _LOGGER.debug(f"self._mapsvgcache_ts: {self._mapsvgcache_ts}")
        _LOGGER.debug(f"self._svg_xPos: {self._svg_xPos}")
        _LOGGER.debug(f"self._svg_yPos: {self._svg_yPos}")

        # User-friendly mower state
        _LOGGER.debug("-self.MowerStateDescription()")
        _LOGGER.debug(
            f"self._mower_state_description = {self._mower_state_description}"
        )

        # split runtime into total
        _LOGGER.debug("-self.RuntimeTotal()")
        _LOGGER.debug(f"self._total_operation: {self._total_operation}")
        _LOGGER.debug(f"self._total_charge: {self._total_charge}")
        _LOGGER.debug(f"self._total_cut: {self._total_cut}")

        # split runtime into session
        _LOGGER.debug("-self.RuntimeSession()")
        _LOGGER.debug(f"self._session_operation: {self._session_operation}")
        _LOGGER.debug(f"self._session_charge: {self._session_charge}")
        _LOGGER.debug(f"self._session_cut: {self._session_cut}")

        # self.getUsers()
        _LOGGER.debug("-self.getUsers")
        _LOGGER.debug(f"self._email: {self._email}")
        _LOGGER.debug(f"self._display_name: {self._display_name}")
        _LOGGER.debug(f"self._language: {self._language}")
        _LOGGER.debug(f"self._country = {self._country}")
        _LOGGER.debug(f"self._optIn = {self._optin}")
        _LOGGER.debug(f"self._optInApp = {self._optinapp}")

        # self.getGenericData()
        _LOGGER.debug("-self.getGenericData")
        # _LOGGER.debug(f"self._alm_sn ignored: {self._alm_sn}")
        _LOGGER.debug(f"self._alm_name: {self._alm_name}")
        _LOGGER.debug(f"self._service_counter: {self._service_counter}")
        _LOGGER.debug(f"self._needs_service: {self._needs_service}")
        _LOGGER.debug(f"self._alm_mode: {self._alm_mode}")
        _LOGGER.debug(f"self._bareToolnumber: {self._bareToolnumber}")
        _LOGGER.debug(f"self._alm_firmware_version: {self._alm_firmware_version}")

        # User-friendly mower model
        _LOGGER.debug("-self._model_description")
        _LOGGER.debug(f"self._model_description = {self._model_description}")

        # self.getOperationalData()
        _LOGGER.debug("-self.getOperatingData")
        _LOGGER.debug(f"battery: {self._battery}")
        _LOGGER.debug(f"garden: {self._garden}")
        _LOGGER.debug(f"hmiKeys: {self._hmikeys}")
        _LOGGER.debug(f"self._battery_percent: {self._battery_percent}")
        _LOGGER.debug(f"self._battery_voltage: {self._battery_voltage}")
        _LOGGER.debug(f"self._battery_cycles: {self._battery_cycles}")
        _LOGGER.debug(f"self._battery_discharge: {self._battery_discharge}")
        _LOGGER.debug(f"self._battery_ambient_temp: {self._battery_ambient_temp}")
        _LOGGER.debug(f"self._battery_temp: {self._battery_temp}")

        # self.getUpdates()
        _LOGGER.debug(f"self._firmware_available: {self._firmware_available}")

        # self.getLastCompletedMow()
        _LOGGER.debug(f"self._last_completed_mow: {self._last_completed_mow}")

        # self.getNextMow()
        _LOGGER.debug(f"self._next_mow: {self._next_mow}")

        # Not updated in the getState API call
        # _LOGGER.debug("Not updated in the getState API call")
        # _LOGGER.debug(f"self._model = {self._model}")

        _LOGGER.debug("-----Show vars end")

    ###########################################################
    ### Updating classes that updates cached data
    ###########################################################

    def getAlerts(self):
        _LOGGER.info("Call getAlerts")
        complete_url = "alerts"
        # _LOGGER.debug(">>>API Call: " + complete_url)
        tmp_json = self.get(complete_url)
        self._alerts = tmp_json
        _LOGGER.debug("--- getAlerts: end")
        return tmp_json

    def getGenericData(self):
        _LOGGER.info("Call getGenericData")
        complete_url = "alms/" + self._serial
        # _LOGGER.debug(f">>>API call: {complete_url}")
        tmp_json = self.get(complete_url)
        self._alm_name = tmp_json.get("alm_name")
        _LOGGER.debug(f"self._alm_name: {self._alm_name}")
        self._service_counter = tmp_json.get("service_counter")
        _LOGGER.debug(f"self._service_counter: {self._service_counter}")
        self._needs_service = tmp_json.get("needs_service")
        _LOGGER.debug(f"self._needs_service: {self._needs_service}")
        self._alm_mode = tmp_json.get("alm_mode")
        _LOGGER.debug(f"self._alm_mode: {self._alm_mode}")
        self._bareToolnumber = tmp_json.get("bareToolnumber")
        _LOGGER.debug(f"self._bareToolnumber: {self._bareToolnumber}")
        self._alm_firmware_version = tmp_json.get("alm_firmware_version")
        _LOGGER.debug(f"self._alm_firmware_version: {self._alm_firmware_version}")
        _LOGGER.debug("--- getGenericData: end")
        return tmp_json

    def getLastCompletedMow(self):
        _LOGGER.info("Call getLastCompletedMow")
        complete_url = "alms/" + self._serial + "/predictive/lastcutting"
        _LOGGER.debug("Complete URL: " + complete_url)
        tmp_json = self.get(complete_url)
        tmp_datetime = tmp_json["last_mowed"]
        self._last_completed_mow = self.convert_bosch_datetime(tmp_datetime)
        _LOGGER.debug(f"tmp_json = {tmp_json}")
        _LOGGER.debug(f"last_completed_mow = {self._last_completed_mow}")
        _LOGGER.debug("--- getLastCompletedMow: end")
        return tmp_json

    def getLocation(self):
        _LOGGER.info("Call getLocation")
        complete_url = "alms/" + self._serial + "/predictive/location"
        Runtime_temp = self.get(complete_url)
        value = Runtime_temp
        _LOGGER.debug("--- getLocation: end")
        return value

    def getNextMow(self):
        _LOGGER.info("Call getNextMow")
        complete_url = "alms/" + self._serial + "/predictive/nextcutting"
        _LOGGER.debug("Complete URL: " + complete_url)
        tmp_json = self.get(complete_url)
        try:
            tmp_datetime = tmp_json["mow_next"]
        except:
            self._next_mow = "None"
        else:
            self._next_mow = self.ConvertBoschDateTime(tmp_datetime)
        _LOGGER.debug(f"NextMow = {self._next_mow}")
        _LOGGER.debug("--- getNextMow: end")
        return tmp_json

    def getOperatingData(self):
        _LOGGER.info("Call getOperatingData")
        complete_url = "alms/" + self._serial + "/operatingData"
        _LOGGER.debug(">>>API Call: " + complete_url)
        tmp_json = self.get(complete_url)
        ### Dont pay attention to runtime values as they are collected in the STATE call also
        if tmp_json:
            _LOGGER.debug(f"runtime: {tmp_json.get('runtime')}")
            self._battery = tmp_json.get("battery")
            _LOGGER.debug(f"battery: {self._battery}")
            self._garden = tmp_json.get("garden")
            _LOGGER.debug(f"garden: {self._garden}")
            self._hmikeys = tmp_json.get("hmiKeys")
            _LOGGER.debug(f"hmiKeys: {self._hmikeys}")
            self._online = True
            self._offline = 0
            _LOGGER.debug(f"Online: {self._online} - Offline: {self._offline}")
            _LOGGER.debug("--- getOperatingData: end")
            return tmp_json
        else:
            self._offline += 1
            self._online = False
            _LOGGER.warning("Mower offline!!!")
            _LOGGER.debug(
                "Online: " + str(self._online) + " - Offline: " + str(self._offline)
            )
            _LOGGER.debug("self._online: " + str(self._online))
            _LOGGER.debug("--- getOperatingData: end")
            return None

    def getState(self):
        _LOGGER.info("Call getState")
        complete_url = "alms/" + self._serial + "/state"
        _LOGGER.debug("URL: " + complete_url)
        tmp_json = self.get(complete_url)
        if tmp_json:
            self._mower_state = tmp_json.get("state")
            _LOGGER.debug(f"self._mower_state: {self._mower_state}")
            self._map_update_available = tmp_json.get("map_update_available")
            _LOGGER.debug(f"self._map_update_available: {self._map_update_available}")
            self._mowed = tmp_json.get("mowed")
            _LOGGER.debug(f"self._mowed: {self._mowed}")
            self._mowmode = tmp_json.get("mowmode")
            _LOGGER.debug(f"self._mowmode: {self._mowmode}")
            self._xpos = tmp_json.get("xPos")
            _LOGGER.debug(f"self._xPos: {self._xpos}")
            self._ypos = tmp_json.get("yPos")
            _LOGGER.debug(f"self._yPos: {self._ypos}")
            self._runtime = tmp_json.get("runtime")
            _LOGGER.debug(f"self._runtime: {self._runtime}")
            self._mapsvgcache_ts = tmp_json.get("mapsvgcache_ts")
            _LOGGER.debug(f"self._mapsvgcache_ts: {self._mapsvgcache_ts}")
            self._svg_xPos = tmp_json.get("svg_xPos")
            _LOGGER.debug(f"self._svg_xPos: {self._svg_xPos}")
            self._svg_yPos = tmp_json.get("svg_yPos")
            _LOGGER.debug(f"self._svg_yPos: {self._svg_yPos}")
            _LOGGER.debug("--- getState end")
            return tmp_json
        else:
            _LOGGER.error("--- getState gave no value when fetching form API")
            _LOGGER.debug("--- getState end")
            return None

    def getForcedState(self):
        # Forces server to also update the location!
        _LOGGER.info("Call getForcedState")
        complete_url = "alms/" + self._serial + "/state" + "?forceRefresh=true"
        _LOGGER.debug("URL: " + complete_url)
        tmp_json = self.get(complete_url)
        self._mower_state = tmp_json.get("state")
        _LOGGER.debug(f"self._mower_state: {self._mower_state}")
        self._map_update_available = tmp_json.get("map_update_available")
        _LOGGER.debug(f"self._map_update_available: {self._map_update_available}")
        self._mowed = tmp_json.get("mowed")
        _LOGGER.debug(f"self._mowed: {self._mowed}")
        self._mowmode = tmp_json.get("mowmode")
        _LOGGER.debug(f"self._mowmode: {self._mowmode}")
        self._xpos = tmp_json.get("xPos")
        _LOGGER.debug(f"self._xPos: {self._xpos}")
        self._ypos = tmp_json.get("yPos")
        _LOGGER.debug(f"self._yPos: {self._ypos}")
        self._runtime = tmp_json.get("runtime")
        _LOGGER.debug(f"self._runtime: {self._runtime}")
        self._mapsvgcache_ts = tmp_json.get("mapsvgcache_ts")
        _LOGGER.debug(f"self._mapsvgcache_ts: {self._mapsvgcache_ts}")
        self._svg_xPos = tmp_json.get("svg_xPos")
        _LOGGER.debug(f"self._svg_xPos: {self._svg_xPos}")
        self._svg_yPos = tmp_json.get("svg_yPos")
        _LOGGER.debug(f"self._svg_yPos: {self._svg_yPos}")
        return tmp_json
        _LOGGER.debug("--- getForcedState end")

    def getLongpollState(self, timeout):
        # The server attempts to "hold open" (not immediately reply to) each HTTP request, responding only when there are events to deliver.
        _LOGGER.info("Call getLongpollState")
        laststate = 0
        if self._mower_state != "":  # check if state variable is already available
            laststate = self._mower_state
        complete_url = (
            "alms/"
            + self._serial
            + "/state"
            + "?longpoll=true&timeout="
            + str(timeout)
            + "&last="
            + str(laststate)
        )  # server will wait with answering until state data has updated values
        _LOGGER.debug("URL: " + complete_url)
        tmp_json = self.get(complete_url, timeout=timeout + 10)
        if tmp_json != None:
            # Longpoll requests do only provide a limited set of status variables
            if tmp_json.get("state") != None:
                self._mower_state = tmp_json.get("state")
                _LOGGER.debug(f"self._mower_state: {self._mower_state}")
            if tmp_json.get("map_update_available") != None:
                self._map_update_available = tmp_json.get("map_update_available")
                _LOGGER.debug(
                    f"self._map_update_available: {self._map_update_available}"
                )
            if tmp_json.get("mowed") != None:
                self._mowed = tmp_json.get("mowed")
                _LOGGER.debug(f"self._mowed: {self._mowed}")
            if tmp_json.get("mowmode") != None:
                self._mowmode = tmp_json.get("mowmode")
                _LOGGER.debug(f"self._mowmode: {self._mowmode}")
            if tmp_json.get("xPos") != None:
                self._xpos = tmp_json.get("xPos")
                _LOGGER.debug(f"self._xPos: {self._xpos}")
            if tmp_json.get("yPos") != None:
                self._ypos = tmp_json.get("yPos")
                _LOGGER.debug(f"self._yPos: {self._ypos}")
            if tmp_json.get("runtime") != None:
                self._runtime = tmp_json.get("runtime")
                _LOGGER.debug(f"self._runtime: {self._runtime}")
            if tmp_json.get("mapsvgcache_ts") != None:
                self._mapsvgcache_ts = tmp_json.get("mapsvgcache_ts")
                _LOGGER.debug(f"self._mapsvgcache_ts: {self._mapsvgcache_ts}")
            if tmp_json.get("svg_xPos") != None:
                self._svg_xPos = tmp_json.get("svg_xPos")
                _LOGGER.debug(f"self._svg_xPos: {self._svg_xPos}")
            if tmp_json.get("svg_yPos") != None:
                self._svg_yPos = tmp_json.get("svg_yPos")
                _LOGGER.debug(f"self._svg_yPos: {self._svg_yPos}")
        return tmp_json
        _LOGGER.debug("--- getLongpollState end")

    def getUpdates(self):
        _LOGGER.info("Call getUpdates")
        if self._online:
            complete_url = "alms/" + self._serial + "/updates"
            tmp_json = self.get(complete_url)
            self._firmware_available = tmp_json.get("available")
            _LOGGER.debug("--- getUpdates: end")
            return tmp_json
        else:
            _LOGGER.warning("Mower offline, no update")
            _LOGGER.debug("--- getUpdates: end")
            return None

    def getUsers(self):
        _LOGGER.info("Call getUsers")
        complete_url = "users/" + self._userid
        _LOGGER.debug(">>>API Call: " + complete_url)
        tmp_json = self.get(complete_url)
        self._email = tmp_json.get("email")
        _LOGGER.debug(f"email = {self._email}")
        self._display_name = tmp_json.get("display_name")
        _LOGGER.debug(f"display_name = {self._display_name}")
        self._language = tmp_json.get("language")
        _LOGGER.debug(f"language = {self._language}")
        self._country = tmp_json.get("country")
        _LOGGER.debug(f"country = {self._country}")
        self._optin = tmp_json.get("optIn")
        _LOGGER.debug(f"optIn = {self._optin}")
        self._optinapp = tmp_json.get("optInApp")
        _LOGGER.debug(f"optInApp = {self._optinapp}")
        _LOGGER.debug(f"Value User = {tmp_json}")
        _LOGGER.debug("--- getUsers: end")
        return tmp_json

    def getNetwork(self):
        _LOGGER.debug("--- getNetwork: start")
        complete_url = "alms/" + self._serial + "/network"
        Runtime_temp = self.get(complete_url)
        value = Runtime_temp
        _LOGGER.debug("--- getNetwork: end")
        return value

    # 8
    def getTest(self):
        # now = datetime.now()
        # _LOGGER.debug(">>>>>>" + now.strftime("%Y-%m-%d %H:%M:%S"))
        _LOGGER.info("Call getTest")
        complete_url = "alms/" + self._serial + "/map"
        _LOGGER.debug("Complete URL: " + complete_url)
        tmp_json = self.get(complete_url)
        _LOGGER.debug(f"tmp_json = {tmp_json}")
        _LOGGER.debug("--- getTest: end")
        return tmp_json

    ###################################################
    ### Functions for getting data from STATE cache

    def MowerState(self):
        if hasattr(self, "_mower_state"):
            return self._mower_state
        else:
            return None

    def MapUpdateAvailable(self):
        if hasattr(self, "_map_update_available"):
            return self._map_update_available
        else:
            return None

    def Mowed(self):
        if hasattr(self, "_mowed"):
            return self._mowed
        else:
            return None

    def MowMode(self):
        if hasattr(self, "_mowmode"):
            return self._mowmode
        else:
            return None

    def XPos(self):
        if hasattr(self, "_xpos"):
            return self._xpos
        else:
            return None

    def YPos(self):
        if hasattr(self, "_ypos"):
            return self._ypos
        else:
            return None

    def Runtime(self):
        if hasattr(self, "_runtime"):
            return self._runtime
        else:
            return None

    def MapSvgCacheTs(self):
        if hasattr(self, "_mapsvgcache_ts"):
            return self._mapsvgcache_ts
        else:
            return None

    def SvgxPos(self):
        if hasattr(self, "_svg_xPos"):
            return self._svg_xPos
        else:
            return None

    def SvgyPos(self):
        if hasattr(self, "_svg_yPos"):
            return self._svg_yPos
        else:
            return None

    ### --- User readable get functions
    def RuntimeTotal(self):
        tmp = self.Runtime()
        if tmp:
            tmp = tmp.get("total")
            if tmp:
                self._total_operation = round(tmp.get("operate") / 100)
                self._total_operation_raw = tmp.get(
                    "operate"
                )  ### Only for testing purposes!
                self._total_charge = round(tmp.get("charge") / 100)
                self._total_cut = round(self._total_operation - self._total_charge)
                return tmp
            else:
                return None
        else:
            return None

    def RuntimeSession(self):
        tmp = self.Runtime()
        if tmp:
            tmp = tmp.get("session")
            if tmp:
                self._session_operation = round(tmp.get("operate"))
                self._session_charge = round(tmp.get("charge"))
                self._session_cut = round(
                    self._session_operation - self._session_charge
                )
                return tmp
            else:
                return None
        return None

    def MowerStateDescription(self):
        if hasattr(self, "_mower_state"):
            if str(self._mower_state) in MOWER_STATE_DESCRIPTION.keys():
                self._mower_state_description = MOWER_STATE_DESCRIPTION.get(
                    str(self._mower_state)
                )
            else:
                _LOGGER.warning(
                    f"Mower State Description not in database! "
                    + str(self._mower_state)
                )
                self._mower_state_description = "Value not in database!"
            return self._mower_state_description
        else:
            return None
        return self._mower_state_description

    def MowerStateDescriptionDetailed(self):
        if hasattr(self, "_mower_state"):
            if str(self._mower_state) in MOWER_STATE_DESCRIPTION_DETAILED.keys():
                self._mower_state_description_detailed = MOWER_STATE_DESCRIPTION_DETAILED.get(
                    str(self._mower_state)
                )
            else:
                _LOGGER.warning(
                    f"Mower State Description Detailed in database! "
                    + str(self._mower_state)
                )
                self._mower_state_description_detailed = "Value not in database!"
            return self._mower_state_description_detailed
        else:
            return None
        return self._mower_state_description_detailed

    ###################################################
    ### Functions for getting data from USERS cache

    def Email(self):
        if hasattr(self, "_email"):
            return self._email
        else:
            return None

    def DisplayName(self):
        if hasattr(self, "_display_name"):
            return self._display_name
        else:
            return None

    def Language(self):
        if hasattr(self, "_language"):
            return self._language
        else:
            return None

    def Country(self):
        if hasattr(self, "_country"):
            return self._country
        else:
            return None

    def OptIn(self):
        if hasattr(self, "_optin"):
            return self._optin
        else:
            return None

    def OptInApp(self):
        if hasattr(self, "_optinapp"):
            return self._optinapp
        else:
            return None

    #########################################################
    ### Functions for getting data from SERIAL API call cache

    def Serial(self):
        return self._serial

    def AlmName(self):
        if hasattr(self, "_alm_name"):
            return self._alm_name
        else:
            return None

    def ServiceCounter(self):
        if hasattr(self, "_service_counter"):
            return self._service_counter
        else:
            return None

    def NeedsService(self):
        if hasattr(self, "_needs_service"):
            return self._needs_service
        else:
            return None

    def AlmMode(self):
        if hasattr(self, "_alm_mode"):
            return self._alm_mode
        else:
            return None

    def BareToolNumber(self):
        if hasattr(self, "_bareToolnumber"):
            return self._bareToolnumber
        else:
            return None

    def AlmFirmwareVersion(self):
        if hasattr(self, "_alm_firmware_version"):
            return self._alm_firmware_version
        else:
            return None

    ### --- User readable get functions

    def ModelDescription(self):
        _LOGGER.info(f"Call ModelDescription")
        if hasattr(self, "_bareToolnumber"):
            _LOGGER.debug(f"bareToolnumber = {self._bareToolnumber}")
            if str(self._bareToolnumber) in MOWER_MODEL_DESCRIPTION.keys():
                self._model_description = MOWER_MODEL_DESCRIPTION.get(
                    str(self._bareToolnumber)
                )
            else:
                _LOGGER.warning(
                    f"Mower Model Description not in database! ModelMumber {self._bareToolnumber}"
                )
                self._model_description = "Value not in database!"
            return self._model_description
        else:
            _LOGGER.error(
                f"ModelDescription bareToolnumber ERROR!!! = {self._bareToolnumber}"
            )
            return None

    def ModelVoltage(self):
        if hasattr(self, "_bareToolnumber"):
            if str(self._bareToolnumber) in MOWER_MODEL_VOLTAGE.keys():
                self._model_voltage = MOWER_MODEL_VOLTAGE.get(str(self._bareToolnumber))
                return self._model_voltage
            else:
                _LOGGER.warning(
                    f"Mower Model Voltage not in database! ModelMumber {self._bareToolnumber}"
                )
        else:
            _LOGGER.error(
                f"ModelVoltage bareToolnumber ERROR!!! = {self._bareToolnumber}"
            )
            return None

    def ModelVoltageMin(self):
        if hasattr(self, "_model_voltage"):
            tmp = self._model_voltage
            self._model_voltage_min = tmp["min"]
            return self._model_voltage_min
        else:
            _LOGGER.error(
                f"ModelVoltageMin bareToolnumber ERROR!!! = {self._bareToolnumber}"
            )
            return None

    def ModelVoltageMax(self):
        if hasattr(self, "_model_voltage"):
            tmp = self._model_voltage
            self._model_voltage_max = tmp["max"]
            return self._model_voltage_max
        else:
            _LOGGER.error(
                f"ModelVoltageMax bareToolnumber ERROR!!! = {self._bareToolnumber}"
            )
            return None

    def MowingModeDescription(self):
        if str(self._alm_mode) in MOWING_MODE_DESCRIPTION.keys():
            self._mowingmode_description = MOWING_MODE_DESCRIPTION.get(
                str(self._alm_mode)
            )
        else:
            _LOGGER.warning(
                f"Mowing Mode Description not in database! Mowing Mode {self._alm_mode}"
            )
            self._mowingmode_description = "Value not in database!"
        return self._mowingmode_description

    ############################################################
    ### Functions for getting data from OPERATING API call cache

    def Battery(self):
        if hasattr(self, "_battery") and (self._battery):
            return self._battery
        else:
            return None

    def BatteryPercent(self):
        tmp = self.Battery()
        if hasattr(self, "_battery") and (self._battery):
            self._battery_percent = tmp.get("percent")
            return self._battery_percent
        else:
            return None

    def BatteryPercentAdjusted(self):
        tmp = self.Battery()
        if (
            hasattr(self, "_battery")
            and (self._battery)
            and (hasattr(self, "_model_voltage_max"))
            and (hasattr(self, "_model_voltage_min"))
        ):
            starttemp = int(self._battery_percent)
            self._battery_percent_adjusted = round(
                (int(starttemp) - int(self._model_voltage_min))
                / ((int(self._model_voltage_max) - int(self._model_voltage_min)) / 100)
            )
            return self._battery_percent_adjusted
        else:
            _LOGGER.error(f"BatteryPerCentAdjusted ERROR!")
            return None

    def BatteryVoltage(self):
        tmp = self.Battery()
        if hasattr(self, "_battery") and (self._battery):
            self._battery_voltage = tmp.get("voltage")
            return self._battery_voltage
        else:
            return None

    def BatteryCycles(self):
        tmp = self.Battery()
        if hasattr(self, "_battery") and (self._battery):
            self._battery_cycles = tmp.get("cycles")
            return self._battery_cycles
        else:
            return None

    def BatteryDischarge(self):
        tmp = self.Battery()
        if hasattr(self, "_battery") and (self._battery):
            self._battery_discharge = tmp.get("discharge")
            return self._battery_discharge
        else:
            return None

    def BatteryAmbientTemp(self):
        tmp = self.Battery()
        if hasattr(self, "_battery") and (self._battery):
            self._battery_ambient_temp = tmp.get("ambient_temp")
            return self._battery_ambient_temp
        else:
            return None

    def BatteryTemp(self):
        tmp = self.Battery()
        if hasattr(self, "_battery") and (self._battery):
            self._battery_temp = tmp.get("battery_temp")
            return self._battery_temp
        else:
            return None

    def Garden(self):
        if hasattr(self, "_garden"):
            return self._garden
        else:
            return None

    def HmiKeys(self):
        if hasattr(self, "_hmikeys"):
            return self._hmikeys
        else:
            return None

    ############################################################
    ### Functions for getting data from UPDATES API call cache

    def FirmwareAvailable(self):
        if hasattr(self, "_firmware_available"):
            return self._firmware_available
        else:
            return None

    ############################################################
    ### Functions for getting data from NEXTCUTTING API call cache

    def NextMow(self):
        return self._next_mow

    ############################################################
    ### Functions for getting data from LASTCUTTING API call cache

    def LastCompletedMow(self):
        return self._last_completed_mow

    ############################################################
    ### Functions for getting data from ALERTS API call cache

    def AlertsDescription(self):
        alerts = self._alerts
        tmp_cnt = 0
        for alert in alerts:
            tmp_cnt += 1
            tmp_code = alert["error_code"]
            if tmp_cnt == 1:
                self._alert1_id = alert["alert_id"]
                self._alert1_time = self.ConvertBoschDateTime(alert["date"])
                self._alert1_error = alert["error_code"]
                self._alert1_friendly_description = self.FriendlyAlertErrorCode(
                    tmp_code
                )
                self._alert2_id = None
                self._alert2_time = None
                self._alert2_error = None
                self._alert2_friendly_description = None
                self._alert3_id = None
                self._alert3_name = None
                self._alert3_error = None
                self._alert3_friendly_description = None
            if tmp_cnt == 2:
                self._alert2_id = alert["alert_id"]
                self._alert2_time = self.ConvertBoschDateTime(alert["date"])
                self._alert2_error = alert["error_code"]
                self._alert2_friendly_description = self.FriendlyAlertErrorCode(
                    tmp_code
                )
                self._alert3_time = None
                self._alert3_error = None
                self._alert3_friendly_description = None
            if tmp_cnt == 3:
                self._alert3_id = alert["alert_id"]
                self._alert3_time = self.ConvertBoschDateTime(alert["date"])
                self._alert3_error = alert["error_code"]
                self._alert3_friendly_description = self.FriendlyAlertErrorCode(
                    tmp_code
                )
        return self._alerts

    def FriendlyAlertErrorCode(self, tmp):
        tmp_val = tmp

        if str(tmp_val) in ALERT_ERROR_CODE.keys():
            alert_description = ALERT_ERROR_CODE[tmp_val]
        else:
            _LOGGER.warning(
                f"Alert Error Code not in database! Alert Code {str(tmp_val)}"
            )
            alert_description = "Not in database!"
        return alert_description

    def ConvertBoschDateTime(self, boshdatetime):
        return boshdatetime[0:10] + " " + boshdatetime[11:16]

    ### --- User readable get functions

    def AlertsCount(self):
        self._alerts_count = len(self._alerts)
        return self._alerts_count

    #######################################
    ### Sending commands to mower
    #######################################
    def putCommand(self, command):
        _LOGGER.info("postCommand: " + command)
        if command == "mow" or command == "pause" or command == "returnToDock":
            complete_url = "alms/" + self._serial + "/state"
            _LOGGER.debug("complete_url: " + complete_url)
            data = '{"state":"' + command + '"}'
            # data: {"state":"mow"}
            temp = self.put(complete_url, data)
            return temp
        else:
            _LOGGER.warning("postCommand " + command + " not valid!")
            return "Wrong Command!"

    def putMowMode(self, command):
        _LOGGER.info("putMowMode command: " + command)
        if (
            command == "true"
            or command == "false"
            or command == "True"
            or command == "False"
        ):
            complete_url = "alms/" + self._serial + "/predictive"
            _LOGGER.debug("complete_url: " + complete_url)
            # command2 = " "enabled": enable }"
            data = '{"enabled":"' + command + '"}'
            temp = self.put(complete_url, data)
            return temp
        else:
            _LOGGER.warning("postCommand " + command + " not valid!")
            return "Wrong Command!"

    def putPredictiveCal(self, command):
        _LOGGER.info("postCommand: " + command)
        complete_url = "alms/" + self._serial + "/predictive/calendar"
        _LOGGER.debug("complete_url: " + complete_url)
        # command2 = " "enabled": enable }"
        data = '{"enabled":"enable"}'
        data = "{'sel_cal': 1, 'cals': [{'cal': 1, 'days': [{'day': 0, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 8, 'EnMin': 0}, {'En': True, 'StHr': 20, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}, {'day': 1, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 8, 'EnMin': 0}, {'En': True, 'StHr': 20, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}, {'day': 2, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 8, 'EnMin': 0}, {'En': True, 'StHr': 20, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}, {'day': 3, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 8, 'EnMin': 0}, {'En': True, 'StHr': 20, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}, {'day': 4, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 8, 'EnMin': 0}, {'En': True, 'StHr': 20, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}, {'day': 5, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 8, 'EnMin': 0}, {'En': True, 'StHr': 20, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}, {'day': 6, 'slots': [{'En': True, 'StHr': 0, 'StMin': 0, 'EnHr': 23, 'EnMin': 59}]}]}]}"
        temp = self.put(complete_url, data)
        return temp

    ###
    # Not properly implemented yet
    ###

    def getPredictiveCalendar(self):
        _LOGGER.info("Call getPredictveCalendar")
        complete_url = "alms/" + self._serial + "/predictive/calendar"
        Runtime_temp = self.get(complete_url)
        value = Runtime_temp
        _LOGGER.debug("--- getPredictveCalendar: end")
        return value

    def getUserAdjustment(self):
        # No idea what this does? Maybe adjust level settings for SmartMowing
        _LOGGER.info("Call getUserAdjustment")
        complete_url = "alms/" + self._serial + "/predictive/useradjustment"
        Runtime_temp = self.get(complete_url)
        value = Runtime_temp
        _LOGGER.debug("--- getUserAdjustment: end")
        return value["user_adjustment"]

    #    def getCalendar(self):
    #        _LOGGER.debug("---")
    #        _LOGGER.debug("getCalendar")
    #        complete_url = 'alms/' + self._serial + '/calendar'
    #        Runtime_temp = self.get(complete_url)
    #        value = Runtime_temp
    #        return value

    #    def getSecurity(self):
    #        _LOGGER.debug("---")
    #        _LOGGER.debug("getSecurity")
    #        complete_url = 'alms/' + self._serial + '/security'
    #        Runtime_temp = self.get(complete_url)
    #        value = Runtime_temp
    #        return value

    #    def getAutomaticUpdate(self):
    #        _LOGGER.debug("---")
    #        _LOGGER.debug("getAutomaticUpdate")
    #        complete_url = 'alms/' + self._serial + '/automaticUpdate'
    #        Runtime_temp = self.get(complete_url)
    #        value = Runtime_temp
    #        return value

    #    def getMap(self):
    #        _LOGGER.debug("---")
    #        print("getMap (Not implemented yet")
    #        #complete_url = 'alms/' + self._serial + '/map'
    #        #Runtime_temp = self.get(complete_url)
    #        #value = Runtime_temp
    #        value = "error"
    #        return value

    ##########################################################################
    ### Basics for API calls
    def get(self, method, timeout=30):
        """Send a GET request and return the response as a dict."""
        _LOGGER.info("Call GET")
        logindata = json.loads(self._login_session.content)
        contextId = logindata["contextId"]
        _LOGGER.debug("      ContextID: " + contextId)
        headers = {CONTENT_TYPE: CONTENT_TYPE_JSON, "x-im-context-id": contextId}
        url = self._api_url + method
        _LOGGER.debug("      >>>API CALL: " + url)
        ###
        ### Test with iteration
        ###
        try_call = 0
        answer = False
        while (try_call < 5) and (answer == False):
            try_call = try_call + 1
            _LOGGER.debug("      GET try " + str(try_call))
            try:
                response = requests.get(url, headers=headers, timeout=timeout)
            except requests.exceptions.Timeout:
                _LOGGER.error("      Failed to update Indego status. Timeout!")
                return None
            except requests.exceptions.TooManyRedirects:
                _LOGGER.error(
                    "      Failed to update Indego status. Too many redirects"
                )
                return None
            except requests.exceptions.RequestException as e:
                _LOGGER.error("      Failed to update Indego status. Error: " + str(e))
                return None
            else:
                _LOGGER.debug("      HTTP Status code: " + str(response.status_code))
                if response.status_code == 504:
                    _LOGGER.error(
                        "      server backend did not have an update before timeout"
                    )
                    return None
                elif response.status_code != 200:
                    _LOGGER.error("      need to call login again")
                    self.login()
                    logindata = json.loads(self._login_session.content)
                    contextId = logindata["contextId"]
                    # return
                else:
                    _LOGGER.debug("      Json:" + str(response.json()))
                    response.raise_for_status()
                    _LOGGER.debug("   --- GET: end")
                    answer = True
                    return response.json()

        if try_call >= 5:
            _LOGGER.warning("Tried 5 times to get data but didnt succeed")
            _LOGGER.warning(
                "Do you have the correct username, password and serial in configuration.yaml???"
            )
        _LOGGER.debug("   --- GET: end")

    def put(self, url, data):
        """Send a PUT request and return the response as a dict."""
        _LOGGER.info("Call PUT")
        logindata = json.loads(self._login_session.content)
        contextId = logindata["contextId"]
        headers = {CONTENT_TYPE: CONTENT_TYPE_JSON, "x-im-context-id": contextId}
        url = self._api_url + url
        _LOGGER.debug("      >>>API CALL: " + url)
        _LOGGER.debug("      headers: " + str(headers))
        _LOGGER.debug("      data: " + str(data))
        try:
            response = requests.put(url, headers=headers, data=data, timeout=30)
        except requests.exceptions.Timeout:
            _LOGGER.error("      Failed to send data to Indego mower. Timeout!")
            return None
        except requests.exceptions.TooManyRedirects:
            _LOGGER.error(
                "      Failed to send data to Indego mower. Too many redirects"
            )
            return None
        except requests.exceptions.RequestException as e:
            _LOGGER.error(
                "      Failed to send data to Indego mower. Error status: " + str(e)
            )
            return None
        else:
            _LOGGER.error("      HTTP Status code: " + str(response.status_code))
            if response.status_code != 200:
                if response.status_code == 400:
                    _LOGGER.error(
                        "      >>>>>>>>>>>>>>>>>> Bad request! <<<<<<<<<<<<<<<<<<<<"
                    )
                elif response.status_code == 404:
                    _LOGGER.error(
                        "      >>>>>>>>>>>>>>>>>> Page not found! <<<<<<<<<<<<<<<<<<<<"
                    )
                elif response.status_code == 500:
                    _LOGGER.error(
                        "      >>>>>>>>>>>>>>>>>> Internal server error! <<<<<<<<<<<<<<<<<<<<"
                    )
                elif response.status_code == 401:
                    _LOGGER.error(
                        "      >>>>>>>>>>>>>>>>>> Unauthorized! <<<<<<<<<<<<<<<<<<<<"
                    )
                    _LOGGER.error("      need to call login again")
                    self.login()
                else:
                    _LOGGER.error(
                        "      >>>>>>>>>>>>>>>>>> Unexpected error! <<<<<<<<<<<<<<<<<<<<"
                    )
                    _LOGGER.error("      need to call login again")
                    self.login()
                return str(response.status_code)
            else:
                _LOGGER.error("      Json:" + str(response))
                _LOGGER.error("   --- GET: end")
                if response.status_code == 200:
                    return "OK"
                else:
                    return str(response.status_code)
        _LOGGER.debug("   --- GET: end")


# End PYPI __init__.py



"""Base class for indego."""
import logging
import typing
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import replace

from .const import ALERT_ERROR_CODE
from .const import CONTENT_TYPE
from .const import CONTENT_TYPE_JSON
from .const import DEFAULT_BODY
from .const import DEFAULT_CALENDAR
from .const import DEFAULT_URL
from .const import MOWER_MODEL_DESCRIPTION
from .const import MOWER_MODEL_VOLTAGE
from .const import MOWER_STATE_DESCRIPTION
from .const import MOWER_STATE_DESCRIPTION_DETAIL
from .const import MOWING_MODE_DESCRIPTION
from .helpers import convert_bosch_datetime
from .states import Alerts
from .states import Battery
from .states import GenericData
from .states import Location
from .states import Network
from .states import OperatingData
from .states import Runtime
from .states import State
from .states import Updates
from .states import Users

_LOGGER = logging.getLogger(__name__)


class IndegoBaseClient(ABC):
    def __init__(
        self,
        username: str,
        password: str,
        serial: str,
        map_filename: str = None,
        api_url: str = DEFAULT_URL,
    ):
        """Abstract class for the Indego Clent, only use the Indego Client or Indego Async Client.

        Args:
            username (str): Username for Indego
            password (str): Password for Indego
            serial (str): Serial number of the mower.

        """
        self._username = username
        self._password = password
        self._serial = serial
        self.map_filename = map_filename
        self._api_url = api_url
        self._logged_in = False
        self._online = False

        self.alerts = [Alerts()]
        self.generic_data = GenericData()
        self.operating_data = OperatingData()
        self.state = State()
        self.updates = Updates()
        self.users = Users()
        self.network = Network()
        self.battery = Battery()
        self.runtime = Runtime()
        self.location = Location()
        self.state_description = None
        self.state_description_detail = None
        self.last_completed_mow = None
        self.next_mow = None
        self.update_available = None
        self.alerts_count = 0

    @abstractmethod
    def update_all(self, force=False):
        pass

    @abstractmethod
    def update_generic_data(self):
        pass

    def _update_generic_data(self, new):
        if new:
            self.generic_data = replace(self.generic_data, **new)
            self._update_battery_percentage_adjusted()

    @abstractmethod
    def update_alerts(self):
        pass

    def _update_alerts(self, new):
        if new:
            self.alerts = [Alerts(**a) for a in new]
            self.alerts_count = len(self.alerts)

    @abstractmethod
    def update_last_completed_mow(self):
        pass

    def _update_last_completed_mow(self, new):
        if new:
            self.last_completed_mow = convert_bosch_datetime(new["last_mowed"])

    @abstractmethod
    def update_location(self):
        pass

    def _update_location(self, new):
        if new:
            self.location = replace(self.location, **new)

    @abstractmethod
    def update_next_mow(self):
        pass

    def _update_next_mow(self, new):
        if new:
            self.next_mow = convert_bosch_datetime(new["mow_next"])

    @abstractmethod
    def update_operating_data(self):
        pass

    def _update_operating_data(self, new):
        if new:
            self._online = True
            self.operating_data = replace(self.operating_data, **new)
            self._update_battery_percentage_adjusted()
        else:
            self._online = False

    def _update_battery_percentage_adjusted(self):
        if (
            self.generic_data.model_voltage.min is not None
            and self.operating_data.battery.percent is not None
        ):
            self.operating_data.battery.update_percent_adjusted(
                self.generic_data.model_voltage
            )

    @abstractmethod
    def update_state(self, force=False, longpoll=False, longpoll_timeout=120):
        pass

    def _update_state(self, new):
        #_LOGGER.debug("---Update State")
        if new:
            self.state = replace(self.state, **new)
        self.state_description = MOWER_STATE_DESCRIPTION.get(str(self.state.state))
        self.state_description_detail = MOWER_STATE_DESCRIPTION_DETAIL.get(str(self.state.state))

    @abstractmethod
    def update_updates_available(self):
        pass

    def _update_updates_available(self, new):
        if new:
            self.update_available = new["available"]

    @abstractmethod
    def update_users(self):
        pass

    def _update_users(self, new):
        if new:
            self.users = replace(self.users, **new)

    @abstractmethod
    def update_network(self):
        pass

    def _update_network(self, new):
        if new:
            self.network = replace(self.network, **new)

    @abstractmethod
    def download_map(self, filename=None):
        pass

    @abstractmethod
    def put_command(self, command: str):
        pass

    @abstractmethod
    def put_mow_mode(self, command: typing.Any):
        pass

    @abstractmethod
    def put_predictive_cal(self, calendar: dict = DEFAULT_CALENDAR):
        pass

    @abstractmethod
    def login(self):
        """Login to the Indego API."""
        pass

    def _login(self, login):
        if login:
            _LOGGER.debug("---Call login")
            self._contextid = login["contextId"]
            self._userid = login["userId"]
            self._logged_in = True
        else:
            self._logged_in = False

    @abstractmethod
    def get(self, path: str, timeout: int = 30):
        """Get implemented by the subclasses either synchronously or asynchronously.

        Args:
            path (str): url to call on top of base_url
            timeout (int, optional): Timeout for the api call. Defaults to 30.
        
        """
        pass

    @abstractmethod
    def put(self, path: str, data: dict):
        """Put implemented by the subclasses either synchronously or asynchronously.

        Args:
            path (str): url to call on top of base_url
            data (dict): data to put
        
        """
        pass

    def __repr__(self):
        return f"{self.generic_data.model_description} ({self.generic_data.alm_sn}) owned by {self.users.display_name}. {self.generic_data}, {self.state}, {self.operating_data}, last mowed: {self.last_completed_mow}, next mow: {self.next_mow}, {self.location}, {self.network}, {self.alerts}, map filename: {self.map_filename}, {self.runtime}, {self.battery}, update available: {self.update_available}, State Descr: {self.state_description}."



from dataclasses import dataclass
from dataclasses import is_dataclass
from datetime import datetime
import logging

_LOGGER = logging.getLogger(__name__)


def nested_dataclass(*args, **kwargs):
    """Wraps a nested dataclass object."""

    def wrapper(cls):
        cls = dataclass(cls, **kwargs)
        original_init = cls.__init__

        def __init__(self, *args, **kwargs):
            for name, value in kwargs.items():
                field_type = cls.__annotations__.get(name, None)
                if is_dataclass(field_type) and isinstance(value, dict):
                    new_obj = field_type(**value)
                    kwargs[name] = new_obj
            original_init(self, *args, **kwargs)

        cls.__init__ = __init__
        return cls

    return wrapper(args[0]) if args else wrapper


def convert_bosch_datetime(dt: str = None):
    """Creates a datetime object from the string from Bosch. Checks if a valid number of milliseconds is sent."""
    _LOGGER.debug("Datetime to be parsed: %s", dt)
    if dt:
        plus_index = dt.find("+")
        dot_index = dt.find(".")
        if dot_index > 0 and plus_index > 0:
            diff = plus_index - dot_index
            if 1 <= diff <= 3:
                dt = dt.replace("+", f"{0*(4-diff)}+")
        return datetime.fromisoformat(dt)
    return None
